package ch.zhaw.pm2.racetrack;

import ch.zhaw.pm2.racetrack.strategy.UserMoveStrategy;
import org.junit.jupiter.api.*;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.List;
import static ch.zhaw.pm2.racetrack.Game.NO_WINNER;
import static ch.zhaw.pm2.racetrack.PositionVector.Direction.*;


/**
 * Test for Class Game. The Class is split up in nested classes.
 */
class GameTest {
    private UserInterface userInterface;
    private Game game;
    private Track track;

    private final String TRACK_FILE_PATH = ".\\tracks\\challenge.txt";
    private final int CAR_INDEX_ONE = 0;
    private final int CAR_INDEX_TWO = 1;

    /**
     * This nested Class tests if the game gets initiated correctly.
     */
    @Nested
    @DisplayName("Test correct Setup")
    class Setup {

        /**
         * Sets up getting the Config, Game and Userinterface for the following tests.
         */
        @BeforeEach
        void setup() {
            userInterface = new UserInterface("Test");
            Config config = new Config();
            game = new Game(userInterface, config);
            try {
                track = game.selectTrack(new File(TRACK_FILE_PATH));
            } catch (InvalidTrackFormatException | FileNotFoundException e) {
                e.printStackTrace();
                Assertions.fail();
            }
            game.selectMoveStrategy(track.getCar(CAR_INDEX_ONE), new UserMoveStrategy(new UserInterface("Testing"), CAR_INDEX_ONE, track.getCarId(CAR_INDEX_ONE)));
            game.selectMoveStrategy(track.getCar(CAR_INDEX_TWO), new UserMoveStrategy(new UserInterface("Testing"), CAR_INDEX_TWO, track.getCarId(CAR_INDEX_TWO)));

        }

        /**
         * Tests if car index is correct
         */
        @Test
        void getCurrentCarIndex() {
            Assertions.assertEquals(CAR_INDEX_ONE, game.getCurrentCarIndex());
            game.switchToNextActiveCar();
            Assertions.assertEquals(CAR_INDEX_TWO, game.getCurrentCarIndex());
        }

        /**
         * Checks if CarId matches char given in track file
         */
        @Test
        void getCarId() {
            Assertions.assertEquals('a', game.getCarId(0));
            Assertions.assertEquals('b', game.getCarId(1));
        }

        /**
         * Checks initial car position
         */
        @Test
        void getCarPosition() {
            Assertions.assertEquals(new PositionVector(24, 22), game.getCarPosition(0));
            Assertions.assertEquals(new PositionVector(24, 24), game.getCarPosition(1));
        }

        /**
         * Checks if initial car velocity is 0,0
         */
        @Test
        void getCarVelocity() {
            Assertions.assertEquals(new PositionVector(0, 0), game.getCarVelocity(0));
            Assertions.assertEquals(new PositionVector(0, 0), game.getCarVelocity(1));
        }

        /**
         * Checks initial winner
         */
        @Test
        void getWinner() {
            Assertions.assertEquals(NO_WINNER, game.getWinner());
        }

        /**
         * Checks correct initial state
         */
        @Test
        void onlyOneCarLeft() {
            Assertions.assertFalse(game.onlyOneCarLeft());
        }

        /**
         * Checks if cars are able to move
         */
        @Test
        void carsMoving() {
            Assertions.assertTrue(game.carsMoving());
        }
    }

    /**
     * This nested Class makes basic manipulation after Game init.
     */
    @Nested
    @DisplayName("Basic manipulation")
    class Manipulation {

        @BeforeEach
        void setup() {
            userInterface = new UserInterface("Test");
            Config config = new Config();
            game = new Game(userInterface, config);
            try {
                track = game.selectTrack(new File(TRACK_FILE_PATH));
            } catch (InvalidTrackFormatException | FileNotFoundException e) {
                e.printStackTrace();
                Assertions.fail();
            }
            game.selectMoveStrategy(track.getCar(CAR_INDEX_ONE), new UserMoveStrategy(new UserInterface("Testing"), CAR_INDEX_ONE, track.getCarId(CAR_INDEX_ONE)));
            game.selectMoveStrategy(track.getCar(CAR_INDEX_TWO), new UserMoveStrategy(new UserInterface("Testing"), CAR_INDEX_TWO, track.getCarId(CAR_INDEX_TWO)));

        }

        /**
         * Checks if car does change velocity specified by direction input
         */
        @Test
        void carTurnCorrect() {
            game.doCarTurn(RIGHT);
            Assertions.assertEquals(new PositionVector(1, 0), game.getCarVelocity(0));
        }

        /**
         *  Checks if car crash is set if a car crashes.
         */
        @Test
        void carCrash() {
            game.doCarTurn(PositionVector.Direction.UP);
            Assertions.assertTrue(game.onlyOneCarLeft());
        }
    }

    /**
     * This nested Class tests a play trough and implements a userInterface which pretends to be a real player.
     * At the end of every Test the user interface stays open for 10 more sec to visualize the game.
     */
    @Nested
    @DisplayName("Playtrough")
    class Play {
        private Game game;

        /**
         * This method will start a game instance on the challenge track. Car a uses MoveListStrategy. Car b uses DoNotMoveStrategy
         */
        @Test
        void winner() {
            game = new Game(new Interface("Test",new Integer[]{0,2,0},new PositionVector.Direction[]{RIGHT,
                RIGHT,
                RIGHT,
                NONE,
                NONE,
                NONE,
                NONE,
                UP,
                LEFT,
                LEFT,
                LEFT,
                LEFT,
                UP_LEFT,
                NONE,
                RIGHT,
                RIGHT,
                RIGHT,
                NONE,
                LEFT,
                DOWN_LEFT,
                DOWN_LEFT,
                LEFT,
                LEFT,
                NONE,
                RIGHT,
                NONE,
                DOWN,
                DOWN,
                RIGHT,
                NONE,
                RIGHT,
                DOWN,
                NONE,
                UP_RIGHT,
                RIGHT,
                UP_RIGHT,
                UP_RIGHT,
                RIGHT,
                RIGHT}), new Config());
            game.initPhase();
            Assertions.assertEquals("a",game.gamePhase());
        }

        /**
         * Will start a game instance where car a does crash.
         */
        @Test
        void crashA() {
            game = new Game(new Interface("Test",new Integer[]{0,1,0},new PositionVector.Direction[]{UP}), new Config());
            game.initPhase();
            Assertions.assertEquals("b",game.gamePhase());
        }

        /**
         * Checks that a car cant win before finishing a complete round around the track.
         */
        @Test
        void passFinishLineInWrongDirection() {
            game = new Game(new Interface("Test",new Integer[]{1,0,1},new PositionVector.Direction[]{LEFT,NONE,NONE,RIGHT,RIGHT}), new Config());
            game.initPhase();
            Assertions.assertEquals("a",game.gamePhase());
        }

        /**
         * Does wait 10 seconds before closing the textio.
         */
        @AfterEach
        void cleanUp() {
            try {
                Thread.sleep(10000);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

    }

    /**
     * This Class is used to communicate with the UserInterface. It overrides crucial methods and returns an instruction based on the instructions' data field.
     * To implement the right instructions the user has to be aware of the game sequence.
     */
    private class Interface extends UserInterface {

        private final PositionVector.Direction[] directions;
        private final Integer[] instructions;
        private int pointerDir,pointerInstruction;

        /**
         * Constructor to create a new Interface Object
         * @param welcometext The first Text which will be printed in the UserInterface Window
         * @param instructions list of instructions to simulate userinput
         * @param directions list of Directions to simulate userinput
         */
        public Interface(String welcometext, Integer[] instructions, PositionVector.Direction[] directions) {
            super(welcometext);
            pointerDir = -1;
            pointerInstruction = -1;
            this.instructions = instructions;
            this.directions = directions;
        }

        /**
         * Overwriting Method for testing without userinteraction
         * @param text    Text which is printed before the options are printed. Example: "Select Track file:"
         * @param options List with the options which can be selected.
         * @return the selected Option
         */
        @Override
        public int selectOption(String text, List<String> options) {
            pointerInstruction++;
            return instructions[pointerInstruction];

        }

        /**
         * Overwriting Method for testing without userinteraction
         * @param text The Text which should be printed.
         */
        @Override
        public void printInformation(String text) {
        }

        /**
         * Overwriting Method for testing without userinteraction
         * @param text Output Text
         */
        @Override
        public void quit(String text) {
        }

        /**
         * Overwriting Method for testing without userinteraction
         * @param playingCarIndex the index of the player
         * @param playingCarID    the ID of the player
         * @return the direction to accelerate
         */
        @Override
        public PositionVector.Direction selectDirection(int playingCarIndex, char playingCarID) {
            pointerDir += 1;
            if(pointerDir < directions.length) {
                return directions[pointerDir];
            }
            return NONE;
        }

    }
}
